#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May  9 16:10:03 2021
Modifié le 16 décembre 2021

@author: ajuton, à partir d'éléments du code de Corentin Dancette
"""
#on importe les modules nécessaires
import random 
import time
import numpy as np #module pour le calcul de matrices
from FrozenLake_2022_fenetre import fenetre_apprentissage, espace_de_jeu_apprentissage

#Création de la fenêtre graphique (voir fichier FrozenLake_2022_fenetre.py)
fenetre = fenetre_apprentissage()
espace_jeu = espace_de_jeu_apprentissage(fenetre)

#Définition de l'environnement, états, récompenses, actions
class Game:
    ACTION_H = 0
    ACTION_G = 1
    ACTION_D = 2
    ACTION_B = 3
    #table des actions possibles
    ACTIONS = [ACTION_H, ACTION_G, ACTION_D, ACTION_B]
    #table des noms des actions
    ACTION_NAMES = ["H", "G", "D", "B"]
    #coordonnées des mouvements
    MOUVEMENTS = {
        ACTION_H: (0, -1),
        ACTION_G: (-1, 0),
        ACTION_D: (1, 0),
        ACTION_B: (0, 1)
    }
    #nombre d'actions possibles
    num_actions = len(ACTIONS)

    #intilialisation de la grille avec largeur, hauteur et wrong_action_p
    def __init__(self, n, m):
        self.n = n
        self.m = m
        self.generate_game()

    #renvoie le numéro de la case (entre 0 et n*m-1) si on donne x et y
    def _position_to_id(self, x, y):
        return x + y * self.n

    #génère une grille de départ avec des positions aléatoires pour les acteurs
    #A chaque affectation, on retire la case affectée des cases possibles
    def generate_game(self):
        cases = [(x, y) for x in range(self.n) for y in range(self.m)]
        renard = random.choice(cases)
        cases.remove(renard)
        start = random.choice(cases)
        cases.remove(start)
        carotte = random.choice(cases)
        cases.remove(carotte)
        block = random.choice(cases)
        cases.remove(block)

        self.position_lapin = start
        self.carotte = carotte 
        self.renard = renard 
        self.block = block
        self.counter = 0
        self.start = start
        return self._get_state()
    
    #repositionne le lapin sur la case de départ de la grille et remet le compteur à 0
    def reset(self):
        self.position_lapin = self.start
        self.counter = 0
        return self._get_state()

    def _get_state(self):
        return self._position_to_id(*self.position_lapin)

    #fonction de déplacement : recoit une action et renvoie :
    #le nouvel état de la grille, la récompense, 
    #si c'est terminé et des infos (0 ici)
    #Il est possible d'y ajouter un peu d'aléatoire (les glissades du lapin)
    def move(self, action):
        self.counter += 1

        if action not in self.ACTIONS:
            raise Exception("Invalid action")

        #ajout d'une action aléatoire :
        #le lapin glisse et ne fait pas ce qu'il aurait prévu
        # choice = random.random()
        #if choice < 0.1 :
        #    action = (action + 1) % 4
        #elif choice < 2 * 0.1 :
        #    action = (action - 1) % 4

        #calcul du déplacement prévu -> new_x et new_y        
        d_x, d_y = self.MOUVEMENTS[action]
        x, y = self.position_lapin
        new_x, new_y = x + d_x, y + d_y

        #calcul des conséquences de l'action : 
        #nouvelle position, récompense, fin du jeu, info(inutilisée)
        if self.block == (new_x, new_y):
            return self._get_state(), -1, False, 0
        elif self.renard == (new_x, new_y):
            self.position_lapin = new_x, new_y
            return self._get_state(), -10, True, 0 
        elif self.carotte == (new_x, new_y):
            self.position_lapin = new_x, new_y
            return self._get_state(), 10, True, 0
        elif new_x >= self.n or new_y >= self.m or new_x < 0 or new_y < 0:
            return self._get_state(), -1, False, 0
        elif self.counter > 190:
            self.position_lapin = new_x, new_y
            return self._get_state(), -10, True, 0
        else:
            self.position_lapin = new_x, new_y
            return self._get_state(), -1, False, 0

    #affichage des acteurs à leur nouvelle position
    def print(self,espace_jeu): 
        for i in range(self.n - 1, -1, -1):
            for j in range(self.m):
                if (i, j) ==   self.position_lapin  :
                    espace_jeu.espace.coords(espace_jeu.lapin, (i*100+50, j*100+50))
                elif (i, j) == self.block:
                    espace_jeu.espace.coords(espace_jeu.bloc, (i*100+50, j*100+50))
                elif (i, j) == self.renard:
                    espace_jeu.espace.coords(espace_jeu.renard, (i*100+50, j*100+50))
                elif (i, j) == self.carotte:
                    espace_jeu.espace.coords(espace_jeu.carotte, (i*100+50, j*100+50))
        
#--------------------------------------------------------------------------
#Algorithme d'apprentissage        
states_n = 16
num_episodes = 200

while(True):
    #initialisations
    cumul_reward_list = []
    actions_list = []
    states_list = []
    #Création d'un nouvel environnement aléatoire
    game = Game(4, 4)
    game.print(espace_jeu)
    Q = np.zeros([states_n, game.num_actions])
    espace_jeu.espace.update()
    time.sleep(1.0)
    
    #attente du bouton démarrer
    while(not fenetre.bouton_start):
        espace_jeu.espace.update()
    fenetre.bouton_start=0

    #Démarrage de l'apprentissage sur un nombre d'épisodes fixé
    for i in range(num_episodes):
        #arrêt de l'apprentissage si appui sur démarrer
        if(fenetre.bouton_start):
            fenetre.bouton_start=0
            break
        
        #RAZ de l'environnement au début de chaque épisode
        actions = []
        s = game.reset()
        states = [s]
        cumul_reward = 0
        fin_du_jeu = False
        game.print(espace_jeu)
        espace_jeu.espace.update()
        time.sleep(0.2)
        
        #démarrage de l'apprentissage jusqu'à la fin de l'épisode (Renard ou Carotte trouvés)
        while True:
            #Mise à jour des hyper-paramètres avec les valeurs indiquées par les boutons
            lr=float(fenetre.vitesse_recu.get()) #vitesse d'apprentissage learning rate .85 par défaut
            y=float(fenetre.facteur_actua_recu.get()) #facteur d'actualisation .99 par défaut
            facteur_explo=float(fenetre.facteur_explo_recu.get()) #facteur d'exploration 1 par défaut
            duree_entre_2_tests=float(fenetre.duree_entre_2_tests_recu.get()) #duree_entre_2_tests .5 par défaut
            
            # on choisit une action aléatoire avec une certaine probabilité, qui décroît petit à petit
            #ou on choisit l'action permettant d'espérer la meilleure récompense
            if random.random() < facteur_explo*(1. / (i/10 +1)) : #décroissance à régler 
                a = random.randint(0,game.num_actions-1)
            else :
                a = np.argmax(Q[s,:])
            
            #obtention par l'environnement de l'état suivant et de la récompense
            s1, reward, fin_du_jeu, _ = game.move(a)
            
            #Mise à jour de la table Q
            Q[s, a] = Q[s, a] + lr*(reward + y * np.max(Q[s1,:]) - Q[s, a])
            cumul_reward += reward
            s = s1
            
            #Mise à jour des historiques des actions et états
            actions.append(a)
            states.append(s)
            
            #Mise à jour de kl'affichage
            game.print(espace_jeu)
            espace_jeu.espace.update()
 
            #affichage de la Qtable
            fenetre.affichage_de_la_qtable(Q,game.num_actions,game.ACTION_NAMES)
  
            #si moode auto, on attend duree_entre_2_tests, si mode_manu, on attend le bouton pas_suivant
            if(fenetre.mode_manuel==0):
                time.sleep(duree_entre_2_tests)
            else:
                while fenetre.pas_suivant==0 :
                    espace_jeu.espace.update()
                fenetre.pas_suivant = 0 #une fois pas_suivant pris en compte, on le remet à zéro
            if fin_du_jeu == True:
                break            
            
        #sauvegarde des historiques des états et actions
        states_list.append(states)
        actions_list.append(actions)
        cumul_reward_list.append(cumul_reward)
    
    #A la fin, affichage du score.
    print("Score over time: " +  str(sum(cumul_reward_list[-100:])/100.0))
    espace_jeu.espace.update()
espace_jeu.espace.mainloop()