#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 28 15:34:30 2021

@author: ajuton
"""

import numpy as np
import time
from collections import deque
from Voiture_autonome_2022_fenetre_v2 import fenetre_apprentissage
from Voiture_autonome_2022_voiture_v2 import voiture

def relu(mat):
    return np.multiply(mat,(mat>0))

def relu_derivative(mat):
    return (mat>0)*1

# classe décrivant une couche de réseau de neurones
class NNLayer:
    # initialisation : nombres d'entrées, nombre de neurones (sorties), fonction d'activation
    def __init__(self, input_size, output_size, activation=None, lr = 0.001):
        self.input_size = input_size
        self.output_size = output_size
        #les poids sont initialisés avec une valeur aléatoire
        self.weights = np.random.uniform(low=-0.5, high=0.5, size=(input_size, output_size))
        self.activation_function = activation
        self.lr = lr
        
    # Calcul des sorties d'une couche à partir des entrées (mode "forward")
    def forward(self, inputs, remember_for_backprop=True):
        input_with_bias = np.append(inputs,1) #ajout d'un biais comme une entrée supplémentaire unitaire
        unactivated = np.dot(input_with_bias, self.weights) #calcul de la somme des entrées pondérée par les poids
        output = unactivated 
        if self.activation_function != None: #si il y a une fonction d'activation, on l'applique
            output = self.activation_function(output)
        if remember_for_backprop:
            # on sauvegarde les entrées et les sorties avant activation en vue d'un calcul de gradient     
            self.backward_store_in = input_with_bias
            self.backward_store_out = np.copy(unactivated)     
        return output

    #mise à jour des poids avec le gradient, pondéré par le taux d'apprentissage
    def update_weights(self, gradient):
        self.weights = self.weights - self.lr*gradient
    
    #mise à jour de la couche de neurones à partir du gradient venant de la couche suivante
    def backward(self, gradient_from_above):
        adjusted_mul = gradient_from_above
        # en cas de fonction d'activation ReLu, on calcule le produit de la dérivée de ReLu par le gradient venant de la couche suivante
        if self.activation_function != None:
            adjusted_mul = np.multiply(relu_derivative(self.backward_store_out),gradient_from_above)
        #On calcule alors les nouveaux poids en utilisant le produit précédent
        D_i = np.dot( np.transpose(np.reshape(self.backward_store_in, (1,len(self.backward_store_in)))), np.reshape(adjusted_mul, (1,len(adjusted_mul))))
        delta_i = np.dot(adjusted_mul, np.transpose(self.weights))[:-1]
        self.update_weights(D_i)
        return delta_i

#classe décrivant l'agent ayant comme Q-fonction un réseau de neurones pour prendre ses décisions
class RLAgent:
    env = None

    #création de l'agent et de son réseau de neurones avec 3 entrées, 2 couches cachées de 23 neurones et 3 sorties    
    def __init__(self, env):
        self.env = ma_voiture
        self.hidden_size = 24
        self.input_size = 3
        self.output_size = 3
        self.num_hidden_layers = 2
        self.epsilon = 1.0
        self.gamma = 0.95 #taux d'actualisation par défaut
        
        #CREATION DU RESEAU DE NEURONES PRINCIPAL   
        #création de la première couche (on rajoute 1 aux entrées pour le biais)
        self.main_NN = [NNLayer(self.input_size+1, self.hidden_size, activation=relu)]
        #création des autres couches cachées (on rajoute 1 aux entrées pour le biais)
        for i in range(self.num_hidden_layers-1):
            self.main_NN.append(NNLayer(self.hidden_size+1, self.hidden_size, activation=relu))
        #création de la couche de sortie, sans couche d'activation
        self.main_NN.append(NNLayer(self.hidden_size+1, self.output_size))
        #affichage du réseau de neurones
        for i in range(self.num_hidden_layers+1) :
            print(self.main_NN[i].weights)
            print("--------------")

        #CREATION DU RESEAU DE NEURONES CIBLE UTILISE POUR L'OPTIMISATION, à l'identique du principal
        self.target_NN = [NNLayer(self.input_size + 1, self.hidden_size, activation=relu)]
        for i in range(self.num_hidden_layers-1):
            self.target_NN.append(NNLayer(self.hidden_size+1, self.hidden_size, activation=relu))
        self.target_NN.append(NNLayer(self.hidden_size+1, self.output_size))
        #intialisation du réseau cible avec les valeurs du réseau principal
        for i in range(self.num_hidden_layers+1) :
            self.target_NN[i].weights = np.copy(self.main_NN[i].weights)
   
    #exploration ou application de la politique pour choisir une action (simple calcul de la sortie du réseau de neurones)     
    def select_action(self, state):
        values = self.forward(state)
        #choix entre exploitation (la meilleure des values) ou exploration (une valeur de sortie aléatoire)
        if (np.random.random() > self.epsilon):
            return np.argmax(values)
        else:
            return np.random.randint(3)
    
    #calcul de la sortie du réseau de neurones en calculant couche après couche, on active ou non la mémorisation    
    def forward(self, state, remember_for_backprop=True):
        vals = np.copy(state)
        index = 0
        for layer in self.main_NN:
            vals = layer.forward(vals, remember_for_backprop)
            index = index + 1
        return vals
    
    #calcul de la sortie du réseau cible. La mémorisation sera désactivée à l'usage
    def forward_target(self, state, remember_for_backprop=True):
        vals = np.copy(state)
        index = 0
        for layer in self.target_NN:
            vals = layer.forward(vals, remember_for_backprop)
            index = index + 1
        return vals
    
    #entrainement à partir de a, s',s, r : calcul des valeurs d'action données par Q et des valeurs d'actions a' données par Qcible
    def train(self, done, action, new_state, state, reward):
        action_values = self.forward(state, remember_for_backprop=True)     
        next_action_values = self.forward_target(new_state, remember_for_backprop=False)
        experimental_values = np.copy(action_values)
        #application de la formule du deep Q-learning avec -100 si crash ou reward sinon.
        if done:
                experimental_values[action] = -100
        else:
                experimental_values[action] = reward + self.gamma*np.max(next_action_values)
        
        #Mise à jour des poids par la propagation du gradient vers les couches amont
        self.backward(action_values, experimental_values)
        #Diminution de epsilon pour l'exploration
        self.epsilon = self.epsilon if self.epsilon < 0.01 else self.epsilon*0.995
        #Diminution du taux d'apprentissage lr
        for layer in self.main_NN:
            layer.lr = layer.lr if layer.lr < 0.0001 else layer.lr*0.995
 
    #Propagagation du gradient vers les couches amont, mise à jour des poids
    def backward(self, calculated_values, target_values): 
        delta = (calculated_values-target_values)
        for layer in reversed(self.main_NN):
            delta = layer.backward(delta)
            

# Création des variables globales, de la voiture, de la fenêtre de l'agent nommé model
NUM_EPISODES = 10000
MAX_TIMESTEPS = 1000# The main program loop
ma_voiture = voiture(300,85,10)
fenetre=fenetre_apprentissage(ma_voiture)
fenetre.ma_fenetre.title('Voiture autonome - Apprentissage par renforcement profond')
fenetre.maj_affichage()
fenetre.learning_rate_recu.set(0.001)
fenetre.facteur_actua_recu.set(0.95)
fenetre.facteur_explo_recu.set(0.99)
model = RLAgent(ma_voiture)
model_best = RLAgent(ma_voiture)
model_best.epsilon = 0 #on enlève l'aléatoire sur model_best qui ne sert que pour l'inférence
cumul_reward = 0
cumul_reward_best = -10000

while(True):
    fenetre.espace_dessin.update()
    
    if fenetre.start :
        fenetre.start= False
        model.epsilon = float(fenetre.facteur_explo_recu.get()) #on remet de l'aléatoire, on repart donc avec beaucoup d'aléatoire
        model.gamma = float(fenetre.facteur_actua_recu.get())
        for i in range(model.num_hidden_layers+1) :
            model.main_NN[i].lr = float(fenetre.learning_rate_recu.get())

        # The main program loop
        for i_episode in range(NUM_EPISODES):
            if(fenetre.start == True) :
                fenetre.start = False
                break

            #positionnement aléatoire de la voiture au départ
            ma_voiture.reset()
            state = ma_voiture.lecture_capteurs(fenetre.image_piste)
            fenetre.maj_affichage()
            cumul_reward = 0
            index_recopie_target_NN = 0
            
            # On commence un pas
            while True :
                index_recopie_target_NN +=1
                #on choisit une action en utilisant la politique ou l'exploration (c'est pris en compte par select_action)
                action = model.select_action(state)
                #on fait un pas et on cumule la récompense
                new_state, reward, done, info = ma_voiture.step(action,fenetre.image_piste)
                cumul_reward += reward
                #affichage si il est activé
                if(fenetre.affichage == True) :
                    fenetre.maj_affichage()
                if(fenetre.start == True) :
                    break
                # On entraine le réseau avec ce pas
                model.train(done, action, new_state, state, reward)
                state = new_state
                
                #A chaque nombre de pas fixé, on recopie le réseau principal dans le réseau cible
                if (index_recopie_target_NN %20 == 0) : #mise à jour de la NN_target
                    for i in range(model.num_hidden_layers+1) :
                        model.target_NN[i].weights = np.copy(model.main_NN[i].weights)
                
                if done:
                    # Lorsque la voiture crash, on affiche le numéro de l'épisode et la récompense
                    print("crash de la voiture épisode i = "+str(i_episode-1)+" cumul_reward = "+str(cumul_reward))
                    #on met à jour la meilleure récompense total et le meilleur réseau
                    if cumul_reward > cumul_reward_best :
                        for i in range(model.num_hidden_layers+1) :
                            model_best.main_NN[i].weights = np.copy(model.main_NN[i].weights)
                        cumul_reward_best = cumul_reward
                    break

    #gestion du rejeu de l'inférence
    if fenetre.rejouer :
        fenetre.rejouer= False
          
        state ,_,_,_ = ma_voiture.step(0,fenetre.image_piste)
        while(True): #tant qu'il n'y a pas d'accident
            action = model_best.select_action(state)
            state ,_,_,_ = ma_voiture.step(action,fenetre.image_piste)
            fenetre.maj_affichage()
            if(state[1]<0.02) : #en cas d'accident on arrête la simulation
                break
            if fenetre.start == True :
                fenetre.start = False
                break
            if fenetre.rejouer :
                fenetre.rejouer= False
                break

fenetre.ma_fenetre.mainloop()

